// FNIRT4D - FMRIB's Non-linear Image Registration Tool
//
// fnirt4D.cpp
//
// Jesper Andersson, FMRIB Image Analysis Group
//
//

#ifdef _OPENMP
#include <omp.h>
#endif

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <string>
#include <vector>
#include <algorithm>
#include <boost/shared_ptr.hpp>
#include "newmat.h"
#include "newmatio.h"
#include "newimage/newimageall.h"
#include "miscmaths/miscmaths.h"
#include "miscmaths/nonlin.h"
#include "basisfield/basisfield.h"
#include "basisfield/splinefield.h"
#include "basisfield/dctfield.h"
#include "matching_points4D.h"
#include "fnirt_costfunctions4D.h"
#include "intensity_mappers4D.h"
#include "fnirtfns4D.h"

#include "warpfns/fnirt_file_writer.h"
#include "warpfns/warpfns.h"

using namespace std;
using namespace NEWMAT;
using namespace NEWIMAGE;
using namespace BASISFIELD;


int main(int   argc,
         char  *argv[])
{
	time_t start;
	time(&start);
	
	// Parse command line and return object that we can 
	// later ask what the user wants us to do.
	boost::shared_ptr<fnirt4D_clp> clp;
	try {
		clp = parse_fnirt_command_line(argc,argv);
	}
	catch (const std::exception& error) {
		cerr << "Error occured when parsing the command line" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	// Read --ref and --in volumes
	boost::shared_ptr<volume4D<float> > ref(new volume4D<float>);
	boost::shared_ptr<volume4D<float> > obj(new volume4D<float>);
	boost::shared_ptr<volume4D<float> > main_obj(new volume4D<float>);
	
	short dType;
	try {
		read_volume4D(*main_obj,clp->Obj(),dType,true);
	}
	catch (const std::exception& error) {
		cerr << "Error occurred when reading --obj file" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	//Take a valid number of DWI volumes
	int nvols = main_obj->tsize();
	if ( clp->GetnumberOfVolumesToUse() <= 0 || clp->GetnumberOfVolumesToUse() > main_obj->tsize() ) {
		clp->SetnumberOfVolumesToUse(main_obj->tsize());
	}
	if ( clp->GetnumberOfVolumesToUse() > 0 && clp->GetnumberOfVolumesToUse() < main_obj->tsize() ){
		nvols = clp->GetnumberOfVolumesToUse(); 
	}
	
	try {
		read_volume4DROI(*ref,clp->Ref(), 0, 0, 0, 0, -1, -1, -1, nvols - 1);
		read_volume4DROI(*obj,clp->Obj(), 0, 0, 0, 0, -1, -1, -1, nvols - 1);
		cout<<"Using "<<nvols<<" volumes\n";
	}
	catch (const std::exception& error) {
		cerr << "Error occurred when reading --ref or --obj file" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}		
	
	//Get the bvals and bvecs if provided
	bool bvalsAndbVectsProvided = false;
	Matrix mainbValues, mainbVectors, bValues, bVectors;	
	if( (clp->Getbvals()).size() > 0){ 
		mainbValues = read_ascii_matrix(clp->Getbvals(), 1, main_obj->tsize());
		if ( mainbValues.SumAbsoluteValue() == 0.0 ) {
			cerr << "Error while reading the bvalues file" << endl;
			exit(EXIT_FAILURE);	
		}
		
		mainbVectors = read_ascii_matrix(clp->Getbvecs(), 3, main_obj->tsize());
		if ( mainbVectors.SumAbsoluteValue() == 0.0 ) {
			cerr << "Error while reading the bvectorss file" << endl;
			exit(EXIT_FAILURE);
		}
		
		bValues = mainbValues.Columns(1, nvols);
		bVectors = mainbVectors.Columns(1, nvols);
		
		bvalsAndbVectsProvided = true;
	}			
	
	if ( !bvalsAndbVectsProvided && clp->applyAngularInterpolation() ) {
		cout<<"Error: Angular Interpolation Selected, but no b-vectors or b-values provided" << endl;
		exit(EXIT_FAILURE);
	}
	
#ifdef _OPENMP
	//Setting up OpenMP
	int numProcs = omp_get_num_procs();
	if ( clp->GetNumberOfProcessors() < numProcs ) numProcs = clp->GetNumberOfProcessors();
	omp_set_num_threads(numProcs);
	cout<<"# threads: "<<clp->GetNumberOfProcessors()<<endl;
#endif

	if ( clp->GetnumberOfVolumesToUse() == 1 ) cout<<"Using only the b0s volume to register the test image (same behavior as fnirt)"<<endl;
	
	// Check for registration to self (sometimes done in TBSS)
	if (trying_to_register_to_self(clp->Ref(),*ref,clp->Obj(),*obj,clp->Affine()) && !clp->applyXformOnly()) {
		try {
			cout<<"Trying to register to itself\n";
			write_self_results(*clp,*ref);
		} 
		catch(const std::exception& error) {
			cerr << "Error occurred when writing results of registration to self" << endl;
			cerr << "Exception thrown with message: " << error.what() << endl; 
		}
		if (clp->Verbose()) cerr << "Input to fnirt was two identical images. Result is unity transform" << endl;
		exit(EXIT_SUCCESS);  // Sort of
	}
	
	cout<<"reference: ("<<ref->xsize()<<", "<<ref->ysize()<<", "<<ref->zsize()<<", "<<ref->tsize()<<")\n";
	cout<<"test     : ("<<obj->xsize()<<", "<<obj->ysize()<<", "<<obj->zsize()<<", "<<obj->tsize()<<")\n";  
	if (!clp->applyXformOnly()) cout<<"Nro levels: "<<clp->NoLevels()<<endl;
	
	// Prepare for fnirting
	boost::shared_ptr<SSD_fnirt4D_CF>       cf;
	boost::shared_ptr<NonlinParam>        nlpar;
	try {
		// Create refmask as combination of explicit and implicit masks.
		boost::shared_ptr<volume<char> > refmask;
		if (clp->UseRefMask(1)) refmask = make_mask(clp->RefMask(),InclusiveMask,*ref,clp->UseImplicitRefMask(),clp->ImplicitRefValue());
		else refmask = make_mask(clp->RefMask(),IgnoreMask,*ref,clp->UseImplicitRefMask(),clp->ImplicitRefValue());
		
		// Create objmask as a combination of explicit and implicit masks.
		boost::shared_ptr<volume<char> > objmask;
		if (clp->UseObjMask(1)) objmask = make_mask(clp->ObjMask(),InclusiveMask,*obj,clp->UseImplicitObjMask(),clp->ImplicitObjValue());
		else objmask = make_mask(clp->ObjMask(),IgnoreMask,*obj,clp->UseImplicitObjMask(),clp->ImplicitObjValue());
		
		// Normalise ref global mean to 100
		for(int t=0; t<ref->tsize(); t++){
			double refmean = spmlike_mean((*ref)[t]);
			(*ref)[t] *= (100.0/refmean);     
		}
		
		// Normalise obj to global mean 100
		for(int t=0; t<obj->tsize(); t++){
			double objmean = spmlike_mean((*obj)[t]);
			(*obj)[t] *= (100.0/objmean);
		}
		
		// Initialise the field that describes the warps.
		std::vector<boost::shared_ptr<basisfield> >    field = init_warpfield(*clp);
		
		// Set up model for intensity mapping
		boost::shared_ptr<IntensityMapper4D>   mymapper = init_intensity_mapper(*clp);
		
		// Create cost-function object
		cf = boost::shared_ptr<SSD_fnirt4D_CF>(new SSD_fnirt4D_CF(*ref,*obj,clp->Affine(),field,mymapper,numProcs));
		
		//New changes
		if ( bvalsAndbVectsProvided ) {
			if ( ref->tsize() > 5 && clp->applyAngularInterpolation() ) {
				cf->EnableAngularInterpolation();
				cout<<"angular interpolation of the DWI image enabled\n";
			}		
			cf->SetbvalsAndbvecs(bValues, bVectors);
		}
		
		// Set masks (if any)
		if (refmask) cf->SetRefMask(*refmask);
		if (objmask) cf->SetObjMask(*objmask);
		
		//Allows applying a given deformation field with Angular Interpolation
		if ( clp->applyXformOnly() ){
			if ( clp->InWarp().size() && clp->ObjOutFname().length() ){
				cout<<"Applying provided deformation field\n";
				
				if ( cf->UseAngularInterpolation() ) main_obj = cf->UpdateObj(true, main_obj, mainbValues, mainbVectors);
				else main_obj = cf->UpdateObj(true, main_obj);
				
				cf->SaveRobj(main_obj, clp->ObjOutFname(), dType);
				
				time_t stop;
				time(&stop);
				cout<<"Time applying the deformation field: "<<difftime(stop,start)/60.0<<"\n";	
				exit(EXIT_SUCCESS);
			}
			else {
				cout<<"inwarp file or image output file name not provided\n";
				exit(EXIT_FAILURE);
			}
		}		
		/////////////////////////////////////////
		
		cf->SetVerbose(clp->Verbose());
		cf->SetRegularisationModel(clp->RegularisationModel());
		cf->SetLambda(clp->Lambda(1));
		cf->SetLevel(1);
		cf->SetIntensityMappingFixed(!clp->EstimateIntensity(1));
		cf->SetHessianPrecision(clp->HessianPrecision());
		if (clp->WeightLambdaBySSD()) cf->WeightLambdaBySSD();
		if (clp->UseRefDeriv()) cf->UseRefDerivs();
		if (clp->Debug()) cf->SetDebug(clp->Debug());
		
		// Smooth and sub-sample
		cf->SmoothRef(clp->RefFWHM(1));
		cf->SmoothObj(clp->ObjFWHM(1));
		cf->SubsampleRef(clp->SubSampling(1));
		
		// Check if we should include point-lists in the warping
		if (clp->RefPointList().length()) {
			PointList         rpl(clp->RefPointList(),clp->Ref());
			rpl.SetAffine(clp->Affine().i());
			PointList         opl(clp->ObjPointList(),clp->Obj());
			MatchingPoints4D    mpl(rpl,opl);
			cf->SetMatchingPoints(mpl);
			cf->SetMatchingPointsLambda(clp->MatchingPointsLambda(1));
		}
		
		// Initialise nonlin parameter object
		NEWMAT::ColumnVector             spar = cf->Par();   // Start guesses for parameters
		nlpar = boost::shared_ptr<NonlinParam>(new NonlinParam(cf->NPar(),clp->MinimisationMethod(),spar));
		set_nlpars(*nlpar);
		nlpar->SetMaxIter(clp->MaxIter(1));
	}
	catch (const std::exception& error) {
		cerr << "Error occurred when preparing to fnirt" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	// And now find parameters for the first level of subsampling
	try {
		// MISCMATHS::NonlinOut status = nonlin(*nlpar,*cf);
		nonlin(*nlpar,*cf);
		if (!constrain_warpfield(*cf,*clp,5)) { // N.B. arbitrary number 5
			std::pair<double,double> range = cf->JacobianRange();
			cout << "Warning, Jacobian not within prescribed range. Prescription is " << clp->JacLowerBound() << " -- " << clp->JacUpperBound();
			cout << " and obtained range is " << range.first << " -- " << range.second << endl;
		}
	}
	catch (const std::exception& error) {
		cerr << "Error occured during estimation at first level of subsampling" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	// Loop over remaining levels of sub-sampling
	try {
		for (int ssl=2; ssl<=clp->NoLevels(); ssl++) {
			if (clp->Verbose()) cout << "***Going to next resolution level***" << endl;
			cf->SetLevel(ssl);
			
			// New mask use
			if (clp->UseRefMask(ssl) != clp->UseRefMask(ssl-1)) {
				boost::shared_ptr<volume<char> > refmask;
				if (clp->UseRefMask(ssl)) refmask = make_mask(clp->RefMask(),InclusiveMask,*ref,clp->UseImplicitRefMask(),clp->ImplicitRefValue());
				else refmask = make_mask(clp->RefMask(),IgnoreMask,*ref,clp->UseImplicitRefMask(),clp->ImplicitRefValue());
				if (refmask) cf->SetRefMask(*refmask);      
			}
			if (clp->UseObjMask(ssl) != clp->UseObjMask(ssl-1)) {
				boost::shared_ptr<volume<char> > objmask;
				if (clp->UseObjMask(1)) objmask = make_mask(clp->ObjMask(),ExclusiveMask,*obj,clp->UseImplicitObjMask(),clp->ImplicitObjValue());
				else objmask = make_mask(clp->ObjMask(),IgnoreMask,*obj,clp->UseImplicitObjMask(),clp->ImplicitObjValue());
				if (objmask) cf->SetObjMask(*objmask);
			}
			
			// New smoothness
			if (clp->RefFWHM(ssl) != clp->RefFWHM(ssl-1)) cf->SmoothRef(clp->RefFWHM(ssl));
			if (clp->ObjFWHM(ssl) != clp->ObjFWHM(ssl-1)) cf->SmoothObj(clp->ObjFWHM(ssl));
			
			// New subsampling
			cout << "Setting subsampling" << endl;
			if (clp->SubSampling(ssl) != clp->SubSampling(ssl-1)) cf->SubsampleRef(clp->SubSampling(ssl));
			cout << "Setting reg mode" << endl;
			cf->SetRegularisationModel(clp->RegularisationModel());
			cout << "Setting lambda" << endl;
			cf->SetLambda(clp->Lambda(ssl));  // New (possibly) lambda
			cf->SetMatchingPointsLambda(clp->MatchingPointsLambda(ssl));
			cf->SetIntensityMappingFixed(!clp->EstimateIntensity(ssl));
			
			// Splash out on brand new nonlin object
			nlpar = boost::shared_ptr<NonlinParam>(new NonlinParam(cf->NPar(),clp->MinimisationMethod(),cf->Par()));
			set_nlpars(*nlpar);
			nlpar->SetMaxIter(clp->MaxIter(ssl));
			
			// MISCMATHS::NonlinOut status = nonlin(*nlpar,*cf);
			nonlin(*nlpar,*cf);
			if (!constrain_warpfield(*cf,*clp,5)) { // N.B. arbitrary number 5
				std::pair<double,double> range = cf->JacobianRange();
				cout << "Warning, Jacobian not within prescribed range. Prescription is " << clp->JacLowerBound() << " -- " << clp->JacUpperBound();
				cout << " and obtained range is " << range.first << " -- " << range.second << endl;
			}
		}
		
		// If we stopped short of full resolution, 
		// up-sample field to full resolution.
		if (clp->SubSampling(clp->NoLevels()) > 1) cf->SubsampleRef(1);
	}
	catch (const std::exception& error) {
		cerr << "Error occured during estimation at subsampling level > 1" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	//Update the full test image
	if ( clp->applyAngularInterpolation() ) {
		cout<<"Updating moving image using Angular and Sinc Interpolation\n";
		cf->SetbvalsAndbvecs(mainbValues, mainbVectors);
		main_obj = cf->UpdateObj(true, main_obj, mainbValues, mainbVectors);
	}
	else {
		cout<<"Updating moving image using Sinc Interpolation\n";
		main_obj = cf->UpdateObj(true, main_obj);
	}
	
	time_t stop1;
	time(&stop1);
	cout<<"Time optimizing the deformation field: "<<difftime(stop1,start)/60.0<<"\n";	
	
	// Save everything we have been asked to save
	try {
		if (clp->CoefFname().length()) cf->SaveDefCoefs(clp->CoefFname());       // Coefficients
		if (clp->FieldFname().length()) cf->SaveDefFields(clp->FieldFname());    // Field
		if (clp->JacFname().length()) cf->SaveJacobian(clp->JacFname());         // Jacobian
		if (clp->RefOutFname().length()) cf->SaveScaledRef(clp->RefOutFname());  // Intensity modulated ref scan
		// Intensity-mapping
		if (clp->IntensityMappingFname().length()) cf->SaveIntensityMapping(clp->IntensityMappingFname());
		// Warped object image
		if (clp->ObjOutFname().length()) {
			cf->SaveRobj(main_obj,clp->ObjOutFname(),dType);                                  
		}
	}
	catch(const std::exception& error) {
		cerr << "Error occured while writing output from fnirt" << endl;
		cerr << "Exception thrown with message: " << error.what() << endl; 
		exit(EXIT_FAILURE);
	}
	
	time_t stop2;
	time(&stop2);
	
	cout<<"Time saving the outputs: "<<difftime(stop2,stop1)/60.0<<"\n";	
	cout<<"total time: "<<difftime(stop2,start)/60.0<<"\n";		
	
	exit(EXIT_SUCCESS);  // R.I.P.
}

