/*  costfns4D.cc

    Copyright (C) 1999-2000 University of Oxford  

    Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    
    LICENCE
    
    FMRIB Software Library, Release 4.0 (c) 2007, The University of
    Oxford (the "Software")
    
    The Software remains the property of the University of Oxford ("the
    University").
    
    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.
    
    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.
    
    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.
    
    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Isis
    Innovation Limited ("Isis"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    innovation@isis.ox.ac.uk quoting reference DE/1112. */

// Interpolation functions
//  Written by Mark Jenkinson  18/2/99

#include <iostream>
#include <cassert>
#include "costfns4D.h"
#include "newmatio.h"
#include "miscmaths/miscmaths.h"
#include "newimage.h"
#include "newimagefns.h"

using namespace MISCMATHS;


// make SAFE_FLIRT the default now (bounds checking on all interpolations)
#define SAFE_FLIRT 1

//Explicit inverse of a 3x3 matrix
bool InverseOfThreeByThreeMatrix(const Matrix& A, Matrix& B) 
{
	//det = a11(a33a22-a32a23)-a21(a33a12-a32a13)+a31(a23a12-a22a13)
	// b11 = (a33a22-a32a23)/det
	// b12 = -(a33a12-a32a13)/det
	// b13 = (a23a12-a22a13)/det
	// b21 = -(a33a21-a31a23)/det
	// b22 = (a33a11-a31a13)/det
	// b23 = -(a23a11-a21a13)/det
	// b31 = (a32a21-a31a22)/det
	// b32 = -(a32a11-a31a12)/det
	// b33 = (a22a11-a21a12)/det
	B = IdentityMatrix(3);
	double det = A(1,1) * ( A(3,3)*A(2,2) - A(3,2)*A(2,3) ) - A(2,1) * ( A(3,3)*A(1,2) - A(3,2)*A(1,3) ) + A(3,1) * ( A(2,3)*A(1,2) - A(2,2)*A(1,3) );  
	if ( det == 0.0 ) {
		return false;
	}
	B(1,1) =  ( A(3,3) * A(2,2) - A(3,2) * A(2,3) ) / det; 
	B(1,2) = -( A(3,3) * A(1,2) - A(3,2) * A(1,3) ) / det;
	B(1,3) =  ( A(2,3) * A(1,2) - A(2,2) * A(1,3) ) / det;
	B(2,1) = -( A(3,3) * A(2,1) - A(3,1) * A(2,3) ) / det;
	B(2,2) =  ( A(3,3) * A(1,1) - A(3,1) * A(1,3) ) / det;
	B(2,3) = -( A(2,3) * A(1,1) - A(2,1) * A(1,3) ) / det;
	B(3,1) =  ( A(3,2) * A(2,1) - A(3,1) * A(2,2) ) / det;
	B(3,2) = -( A(3,2) * A(1,1) - A(3,1) * A(1,2) ) / det;
	B(3,3) =  ( A(2,2) * A(1,1) - A(2,1) * A(1,2) ) / det;
	return true;
}

//Taken from decomp_aff in src/miscmats/miscmats.cc and modified to obtain the angularInterpolation matrix only 
int extract_rotation_aff(const Matrix& affmat, Matrix& rotmat)
{
	// decomposes using the convention: mat = rotmat * skew * scale
	// angles are in radians
	rotmat = IdentityMatrix(3); //In case the algorithm fails, better to return an identity matrix
	ColumnVector x(3), y(3), z(3);
	Matrix aff3(3,3);
	aff3 = affmat.SubMatrix(1,3,1,3);
	x = affmat.Column(1);
	y = affmat.Column(2);
	z = affmat.Column(3);
	float sx, sy, sz, a, b, c;
	sx = norm2(x);
	sy = std::sqrt( dot(y,y) - (Sqr(dot(x,y)) / Sqr(sx)) );
	
	if ( sx == 0.0 || sy == 0.0 ) return -1;
	a = dot(x,y)/(sx*sy);
	
	ColumnVector x0(3), y0(3);
	
	if ( sx > 0.0 ) x0 = x/sx;
	else return -1;
	
	if ( sy > 0.0 ) y0 = y/sy - a*x0;
	else return -1;
	
	sz = std::sqrt(dot(z,z) - Sqr(dot(x0,z)) - Sqr(dot(y0,z)));
	
	if ( sz == 0.0 ) return -1;
	
	b = dot(x0,z)/sz;
	c = dot(y0,z)/sz;
	
	aff3.Column(1) /= sx;
	aff3.Column(2) /= sy;	
	aff3.Column(3) /= sz;
	
	Real skewvals[] = {1,a,b, 0,1,c, 0,0,1}; 
	Matrix skew(3,3), skew_inv(3,3);
	skew  << skewvals;
	
	bool success = InverseOfThreeByThreeMatrix(skew, skew_inv);
	if (success) rotmat = aff3 * skew_inv;
	else return -1;
	
	return 0;
}

//This algorithm implements the Angular interpolation of DWI images proposed in "A Method for Registering Diffusion Weighted Magnetic Resonance Images"
//Xiaodong Tao and James V. Miller, MICCAI 2006, Springer-Verlag 2006
//Introduce a variant to the algorithm using kd-trees
void angularInterpolation(const int& t0,					//current t index on the DWI image
						  kdtree*	kd_tree,	//kd-tree to perform a quick search of the closest neighbors to the target b-vector
						  const volume4D<float>& DWI, 
						  const Matrix& bValues,
						  const Matrix& bVectors,
						  const float& targetbValue,
						  const ColumnVector& targetbVector,
						  volume4D<float>& interpolatedVolume)
{
	//float sigma = 0.005f, dotProduct, d = 0.0f;
	float sigma = 0.05f, dotProduct, d = 0.0f;
	float sumOfWeights = 0.0f;
	
	std::vector<float> queryVector(3);
	for(unsigned int l = 0; l<3; l++){
		queryVector[l] = targetbVector(l+1);
	}	
	
	kdtree_result_vector searchResult;
	kd_tree->n_nearest(queryVector, 8, searchResult);		
	
	std::vector< std::pair<int, float> > indicesAndweights;
	for(unsigned int i=0; i<searchResult.size(); i++){
		int t = searchResult[i].idx + 2;
		if ( fabs( bValues(1,t) - targetbValue ) <= maxb0variability ) { //Compare only on the same q-ball
			dotProduct = fabs( dot(bVectors.SubMatrix(1,3,t,t), targetbVector) );
			if (dotProduct >= 0.999f) { // the bVector and target b Vectors are identical or almost identical, so no need to interpolate
				interpolatedVolume[t0] = DWI[t-1];
				return;
			}
			else { 
				d = std::acos( dotProduct );
				d *= d;
				float weight = std::exp(-0.5 * d/sigma);
				indicesAndweights.push_back( std::make_pair(t, weight) );
				sumOfWeights += weight;
			}			
		}
	}	 
	
	if ( indicesAndweights.size() == 0 ) {  //The b-vals are quite different (different protocols)
		float closest_bval = 0.0;
		float minimumDifference = 1.0e100;
		for(int t=2; t<=bVectors.Ncols(); t++){
			float difference = fabs( bValues(1,t) - targetbValue );
			if (  difference < minimumDifference ) {
				minimumDifference = difference;
				closest_bval = bValues(1,t);
			}
		}
		
		if (closest_bval == 0.0){	//this should not happen frequently, but if it does, use the 4 closest neighbors we can find
			kd_tree->n_nearest(queryVector, 4, searchResult);
		}
		
		//Let's try again
		for(unsigned int i=0; i<searchResult.size(); i++){
			int t = searchResult[i].idx + 2;
			if ( fabs( bValues(1,t) - closest_bval ) <= maxb0variability ) { //Compare only on the same q-ball
				dotProduct = fabs( dot(bVectors.SubMatrix(1,3,t,t), targetbVector) );
				if (dotProduct >= 0.999f) { // the bVector and target b Vectors are identical or almost identical, so no need to interpolate
					interpolatedVolume[t0] = DWI[t-1];
					return;
				}
				else { 
					d = std::acos( dotProduct );
					d *= d;
					float weight = std::exp(-0.5 * d/sigma);
					indicesAndweights.push_back( std::make_pair(t, weight) );
					sumOfWeights += weight;
				}			
			}
		}	 
	}
	
	if ( sumOfWeights > 0.0 ) {
		int t = indicesAndweights[0].first;
		float weight = indicesAndweights[0].second / sumOfWeights;
		interpolatedVolume[t0] = weight * DWI[t-1];			//initialization out of the for loop that avoids initializing with zero, which is expensive
		
		for(unsigned int i=1; i<indicesAndweights.size(); i++){
			t = indicesAndweights[i].first;
			weight = indicesAndweights[i].second / sumOfWeights;
			interpolatedVolume[t0] += weight * DWI[t-1];
		}
	}
	else {		//This should never happen, but if it does, then use the closest b-vect
		kd_tree->n_nearest(queryVector, 1, searchResult);			
		int t = searchResult[0].idx + 2;
		interpolatedVolume[t0] = DWI[t];
		return;
	}
}

void MatchTestToReferenceTsamples(kdtree*	kd_tree,
								  const Matrix& bValuesRef, 
								  const Matrix& bVectorsRef, 
								  const Matrix& bValuesTest, 
								  const Matrix& bVectorsTest, 
								  volume4D<float>& inputv)
{
	int t;
	volume4D<float> resampledTestVolume4D;
	for (t=0; t<bValuesRef.Ncols(); t++) {
		resampledTestVolume4D.addvolume( volume<float>() );
	}
	
	resampledTestVolume4D[0] = inputv[0];
	
#pragma omp parallel for private(t)
	for (t=1; t<bValuesRef.Ncols(); t++) { //The first vector is always (0,0,0) so not need to interpolation here
		angularInterpolation(t, 
							 kd_tree, 
							 inputv, 
							 bValuesTest, 
							 bVectorsTest, 
							 bValuesRef(1, t+1), 
							 bVectorsRef.SubMatrix(1,3,t+1,t+1), 
							 resampledTestVolume4D);
	}
	
	inputv = resampledTestVolume4D;
}

void Costfn4D::applyAngularInterpolation(const Matrix& aff)
{
	Matrix rotationMatrix(3, 3);
	extract_rotation_aff(aff, rotationMatrix);
	Matrix rotated_bVectors = rotationMatrix.t() * bVectorsTest;
	
	int t;
# pragma omp parallel for private(t)
	for (t=1; t<tSamples; t++) { //The first vector is always (0,0,0) so not need to interpolation here
		angularInterpolation(t, 
							 kd_tree, 
							 *testVolume4D, 
							 bValuesTest, 
							 bVectorsTest, 
							 bValuesRef(1, t+1), 
							 rotated_bVectors.SubMatrix(1,3,t+1,t+1), 
							 angularlyInterpolatedTestVolume4D);
	}
	
	for (t=1; t<tSamples; t++) {
		Costfn* newCost;
		if (this->validweights) newCost = new Costfn(vCosts[t]->refvol, angularlyInterpolatedTestVolume4D[t], vCosts[t]->rweight, vCosts[t]->tweight);
		else newCost = new Costfn(vCosts[t]->refvol, angularlyInterpolatedTestVolume4D[t]);
		newCost->set_costfn(this->costtype);
		newCost->set_no_bins(this->no_bins);
		newCost->fuzzyfrac = vCosts[t]->fuzzyfrac;
		newCost->smoothsize = vCosts[t]->smoothsize;
		
		delete vCosts[t];
		vCosts[t] = newCost;
	}
	
	updateGOGTestImage = true;	
}

void Costfn4D::applyAngularInterpolation(const Matrix& aff, volume4D<float>& inputv)
{
	Matrix rotationMatrix(3, 3);
	
	extract_rotation_aff(aff, rotationMatrix);
	Matrix rotated_bVectors = rotationMatrix.t() * bVectorsTest;
	
	int t;
	angularlyInterpolatedTestVolume4D = volume4D<float>();
	for(t=0; t<tSamples; t++){
		angularlyInterpolatedTestVolume4D.addvolume(volume<float>());
	}
	
# pragma omp parallel for private(t)
	for (t=1; t<tSamples; t++) { //The first vector is always (0,0,0) so not need to interpolation here
		angularInterpolation(t, 
							 kd_tree, 
							 inputv, 
							 bValuesTest, 
							 bVectorsTest, 
							 bValuesRef(1, t+1), 
							 rotated_bVectors.SubMatrix(1,3,t+1,t+1), 
							 angularlyInterpolatedTestVolume4D);
	}	
	
# pragma omp parallel for private(t)
	for (t=1; t<tSamples; t++) {
		inputv[t] = angularlyInterpolatedTestVolume4D[t];
	}
}

  
/////////////////////////////////////////////////////////////////////////

//Constructor
Costfn4D::Costfn4D()
{
	testVolume4D = 0;
	angularlyInterpolatedTestVolume4D = volume4D<float>();	
	updateCOGRefImage = true;
	updateGOGTestImage = true;
	angularlyInterpolated = false;
	validweights = false;
	costtype = CorrRatio; 
	no_bins = 0;
	
	refcog = ColumnVector(3);
	testcog = ColumnVector(3);
	tSamples = 0;	
	
	bVectorsRef = Matrix();
	bValuesRef = Matrix();
	bValuesTest = Matrix();
	bVectorsTest = Matrix();
}

//Destructor
Costfn4D::~Costfn4D()
{
	for(int t=0; t<tSamples; t++){
		delete vCosts[t];
	}
}	

//Assumes the reference & test images have equal number of t samples, or one of them contains only a b0 component
//hence, it takes the minimum number of t-samples
void Costfn4D::setNewVolumes(volume4D<float>& refv, volume4D<float>& inputv)
{
	if ( refv.tsize() == inputv.tsize() || refv.tsize() < inputv.tsize() ) tSamples = refv.tsize();
	else tSamples = inputv.tsize();
	
	testVolume4D = &inputv;
	
	if ( vCosts.size() > 0 ) {
		for(unsigned int t=0; t<vCosts.size(); t++){
			delete vCosts[t];
		}
		
		vCosts.clear();
	}
	
	for(int t=0; t<tSamples; t++){
		vCosts.push_back( new Costfn(refv[t], inputv[t]) );
	}
	
	validweights = false;
	updateCOGRefImage = true;
	updateGOGTestImage = true;
}

void Costfn4D::setNewVolumes(volume4D<float>& refv, volume4D<float>& inputv, const volume<float>& refweight, const volume<float>& inweight)
{
	if ( refv.tsize() == inputv.tsize() || refv.tsize() < inputv.tsize() ) tSamples = refv.tsize();
	else tSamples = inputv.tsize();
	
	testVolume4D = &inputv;
	
	if ( vCosts.size() > 0 ) {
		for(unsigned int t=0; t<vCosts.size(); t++){
			delete vCosts[t];
		}
		
		vCosts.clear();
	}
	
	for(int t=0; t<tSamples; t++){
		vCosts.push_back( new Costfn(refv[t], inputv[t], refweight, inweight) );
	}
	
	validweights = true;
	updateCOGRefImage = true;
	updateGOGTestImage = true;	
}

void Costfn4D::setbvalsAndbvects(const Matrix& b_valuesRef, const Matrix& b_vectorsRef, const Matrix& b_valuesTest, const Matrix& b_vectorsTest)
{
	bValuesRef = b_valuesRef;
	bVectorsRef = b_vectorsRef;
	bValuesTest = b_valuesTest;
	bVectorsTest = b_vectorsTest;
	
	treeData.resize(boost::extents[bVectorsTest.Ncols()-1][3]); //Do not inlude the first vector, which is always (0,0,0)
	
	for (int j=1; j<bVectorsTest.Ncols(); j++) {
		for (int i=0; i<3; i++){
			treeData[j-1][i] = bVectorsTest(i+1,j+1);
		}
	}
	
	kd_tree = new kdtree(treeData,false);
	kd_tree->sort_results = true;
}

void Costfn4D::enableAngularInterpolation()
{
	if (bVectorsRef.Ncols() >= 6 && bVectorsTest.Ncols() >= 6) {
		angularlyInterpolated = true;
		angularlyInterpolatedTestVolume4D = volume4D<float>();
		for(int t=0; t<tSamples; t++){
			angularlyInterpolatedTestVolume4D.addvolume(volume<float>());
		}
	}
}

void Costfn4D::disableAngularInterpolation()
{
	angularlyInterpolated = false;
}

void Costfn4D::setSmoothSize(const float& smooth_size)
{
	for(int t=0; t<tSamples; t++){
		vCosts[t]->smoothsize = smooth_size;
	}
}

void Costfn4D::setFuzzyFraction(const float& fuzzy_fraction)
{
	for(int t=0; t<tSamples; t++){
		vCosts[t]->fuzzyfrac = fuzzy_fraction;
	}
}

void Costfn4D::set_costfn(const costfns& costtype)
{
	for(int t=0; t<tSamples; t++) {
		vCosts[t]->set_costfn(costtype);
	}
	this->costtype = costtype;
}

void Costfn4D::set_no_bins(const int& no_bins)
{
	for(int t=0; t<tSamples; t++) {
		vCosts[t]->set_no_bins(no_bins);
	}
	this->no_bins = no_bins;
}


void Costfn4D::resetInputVolume()
{
	for (int t=1; t<tSamples; t++) {
		Costfn* newCost;
		if (this->validweights) newCost = new Costfn(vCosts[t]->refvol, (*testVolume4D)[t], vCosts[t]->rweight, vCosts[t]->tweight);
		else newCost = new Costfn(vCosts[t]->refvol, (*testVolume4D)[t]);
		newCost->set_costfn(this->costtype);
		newCost->set_no_bins(this->no_bins);
		newCost->fuzzyfrac = vCosts[t]->fuzzyfrac;
		newCost->smoothsize = vCosts[t]->smoothsize;
		
		delete vCosts[t];
		vCosts[t] = newCost;
	}
	
	updateGOGTestImage = true;
}

ColumnVector Costfn4D::getCogRefVolume() 
{
	if ( updateCOGRefImage ) {
		refcog = 0.0;
		
		for(int t=0; t<tSamples; t++){
			refcog += (vCosts[t]->refvol).cog("scaled_mm");	
		}
		
		refcog /= tSamples;
		updateCOGRefImage = false;
	}
	
	return refcog;
}

ColumnVector Costfn4D::getCogTestVolume() 
{
	if ( updateGOGTestImage ) {
				
		testcog = 0.0;
		for(int t=0; t<tSamples; t++){
			testcog += (vCosts[t]->testvol).cog("scaled_mm");	
		}
				
		testcog /= tSamples;
		
		updateGOGTestImage = false;
	}
	
	return testcog;
}

Costfn* Costfn4D::GetCostFn(int tValue)
{
	assert( vCosts.size() > 0 && tValue >= 0 && tValue < tSamples );
	return vCosts[tValue];
}

int Costfn4D::count() const
{
	return vCosts[0]->count();
}

//4D cost function
float Costfn4D::cost(const Matrix& affmat) // affmat is voxel to voxel
{			
	if ( tSamples > 1 ) {
		if ( angularlyInterpolated ) applyAngularInterpolation(affmat);
		
		vector<float> costs(tSamples, 0.0);
		int t0;
# pragma omp parallel for private(t0)
		for(t0=0; t0<tSamples; t0++) {
			costs[t0] = vCosts[t0]->cost(affmat);
		}
		
		if ( angularlyInterpolated ) resetInputVolume(); //Return the original test volume to the costs functions
		
		float returnValue = costs[0];
		for(t0=1; t0<tSamples; t0++) {
			returnValue += costs[t0];
		}
		
		returnValue /= tSamples;
		
		return returnValue;	  	  		
	}
	else return vCosts[0]->cost(affmat);
}





