/*  costfns4D.h

    Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    
    LICENCE
    
    FMRIB Software Library, Release 4.0 (c) 2007, The University of
    Oxford (the "Software")
    
    The Software remains the property of the University of Oxford ("the
    University").
    
    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.
    
    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.
    
    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.
    
    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Isis
    Innovation Limited ("Isis"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    innovation@isis.ox.ac.uk quoting reference DE/1112. */


#if !defined(__costfns4D_h)
#define __costfns4D_h

#ifdef _OPENMP
#include <omp.h>
#endif

#include "newimageall.h"
#include "costfns.h"
#include "kdtree.h"

using namespace NEWIMAGE;

const float maxb0variability = 50.0;	//Allows grouping close b-vals within +/- maxb0variability into a single b-val

//Helper functions for the 4D cost function	
int extract_rotation_aff(const Matrix& affmat, Matrix& rotmat);

//Apply angular rotation to whole volumes (linear registration: flirt)
void angularInterpolation(const int& t0,					//current t index on the DWI image
						  kdtree*	kd_tree,	//kd-tree to perform a quick search of the closest neighbors to the target b-vector
						  const volume4D<float>& DWI, 
						  const Matrix& bValues,
						  const Matrix& bVectors,
						  const float& targetbValue,
						  const ColumnVector& targetbVector,
						  volume4D<float>& interpolatedVolume);

void MatchTestToReferenceTsamples(kdtree*	kd_tree,
								  const Matrix& bValuesRef, 
								  const Matrix& bVectorsRef, 
								  const Matrix& bValuesTest, 
								  const Matrix& bVectorsTest, 
								  volume4D<float>& inputv);

class Costfn4D{
	
public:
	// Publicly available calls
	Costfn4D();
	void resetInputVolume();
	void setNewVolumes(volume4D<float>& refv, volume4D<float>& inputv);
	void setNewVolumes(volume4D<float>& refv, volume4D<float>& inputv, const volume<float>& refweight, const volume<float>& inweight);
	
	void setbvalsAndbvects(const Matrix& b_valuesRef, const Matrix& b_vectorsRef, 
						   const Matrix& b_valuesTest, const Matrix& b_vectorsTest);
	void setSmoothSize(const float& smooth_size);
	void setFuzzyFraction(const float& fuzzy_fraction);
	void set_costfn(const costfns& costtype);
	void set_no_bins(const int& no_bins);
	void enableAngularInterpolation();
	void disableAngularInterpolation();
	void applyAngularInterpolation(const Matrix& aff, volume4D<float>& inputv);
	
	Costfn* GetCostFn(int tValue);
	int count() const;
	unsigned int numberOfTimesamples() const {return tSamples;}
	bool isAngularlyInterpolated() const {return angularlyInterpolated;} 
	bool isNotEmpty() const {return (vCosts.size() > 0);}
	
	ColumnVector getCogRefVolume();
	ColumnVector getCogTestVolume();
	~Costfn4D();
	
	// General cost function call
	float cost(const Matrix& affmat);    // affmat is image to image
	
private:
	std::vector<Costfn*>				vCosts;
	volume4D<float>*					testVolume4D;
	volume4D<float>						angularlyInterpolatedTestVolume4D;
	kdtree*							    kd_tree;	
	kdtree_array						treeData;
	Matrix								bVectorsRef;
	Matrix								bValuesRef;
	Matrix								bValuesTest;
	Matrix								bVectorsTest;
	ColumnVector						refcog;
	ColumnVector						testcog;
	costfns							    costtype;
	int									tSamples;
	int									no_bins;
	bool								validweights;
	bool								angularlyInterpolated;
	bool								updateCOGRefImage;
	bool								updateGOGTestImage;
	
	// Prevent default behaviours
	Costfn4D operator=(const Costfn&);
	Costfn4D(const Costfn&);
	
	// Internal functions available
	void applyAngularInterpolation(const Matrix& aff);
};

//////////////////////////////////////////////////////////////////////////

#endif







